using System;
using System.Collections.Generic;

using FTD2XX_NET;
using Libs;

namespace RMT {
	public class FTDIUsage: PicBoard {
		public FTDI FTDI = null;
		public Boolean IsRS232 = false;
		public CmdProc _cmdProc;

		public FTDIUsage(CmdProc cmdProc) {
			this._cmdProc = cmdProc;
		}

		public override Boolean IsClosed() {
			return (this.FTDI == null || this.FTDI.IsOpen == false);
		}

		public override void Open(String description, StringList a, ref String errorText, uint baudRate) {
			FTDI device = new FTDI();
			FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;

			StringList descriptions = StringList.Split(description, ";");
			// Get the number of device attached
			UInt32 deviceCount = 0;
			ftStatus = device.GetNumberOfDevices(ref deviceCount);
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.GetNumberOfDevices failed.";
				this.FTDI = null;
				return;
			}
			// Display device count on form
			a.Add("DeviceCount: " + deviceCount.ToString());
			if (deviceCount == 0) {
				errorText = "No FTDI device found.";
				this.FTDI = null;
				return;
			}

			// Allocate storage for device info list
			FTDI.FT_DEVICE_INFO_NODE[] ftdiDeviceList = new FTDI.FT_DEVICE_INFO_NODE[deviceCount];
			// Populate our device list
			ftStatus = device.GetDeviceList(ftdiDeviceList);
			this.IsRS232 = (baudRate > 0);
			UInt32 index = 10000;
			if (ftStatus == FTDI.FT_STATUS.FT_OK) {
				for (UInt32 i = 0; i < deviceCount; i++) {
					//a.Add("Device Index: " + i.ToString());
					//a.Add("  Flags: " + String.Format("{0:x}", ftdiDeviceList[i].Flags));
					//a.Add("  Type: " + ftdiDeviceList[i].Type.ToString());
					//a.Add("  ID: " + String.Format("{0:x}", ftdiDeviceList[i].ID));
					//a.Add("  Location ID: " + String.Format("{0:x}", ftdiDeviceList[i].LocId));
					//a.Add("  Serial Number: " + ftdiDeviceList[i].SerialNumber.ToString());
					a.Add(i.ToString() + ". Description: " + ftdiDeviceList[i].Description.ToString());
					if (descriptions.Found(ftdiDeviceList[i].Description)) {
						index = i;
					}
				}
			}
			else {
				errorText = "FTDI.GetDeviceList failed.";
				this.FTDI = null;
				return;
			}

			if (String.IsNullOrEmpty(description)) {
				errorText = "Description required.";
				this.FTDI = null;
				return;
			}

			if (index == 10000) {
				errorText = "Device by description not found.";
				this.FTDI = null;
				return;
			}

			// Open first device in our list by serial number
			ftStatus = device.OpenByIndex(index);
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.OpenByIndex failed.";
				this.FTDI = null;
				return;
			}

			// Reset device
			ftStatus = device.ResetDevice();
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.ResetDevice failed.";
				this.FTDI = null;
				return;
			}

			if (this.IsRS232) {
				ftStatus = device.SetLatency(1);
				if (ftStatus != FTDI.FT_STATUS.FT_OK) {
					errorText = "FTDI.SetLatency failed.";
					this.FTDI = null;
					return;
				}
			}
			//else {
			//  if (descriptions.Found("Dual RS232-HS A")) {
			//    ftStatus = device.SetBitMode(0x00, FTD2XX_NET.FTDI.FT_BIT_MODES.FT_BIT_MODE_SYNC_FIFO);
			//    if (ftStatus != FTDI.FT_STATUS.FT_OK) {
			//      errorText = "FTDI.SetBitMode failed.";
			//      this.FTDI = null;
			//      return;
			//    }
			//  }
			//}

			ftStatus = device.SetTimeouts(100, 100);
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.SetTimeouts failed.";
				this.FTDI = null;
				return;
			}

			errorText = "";
			this.FTDI = device;
			//if (this._isRS232) { FTDI_IO._ftdi.SetBaudRate(400000); }			
			//if (this._isRS232) { FTDI_IO._ftdi.SetBaudRate(115200); }
			//if (this._isRS232) { FTDI_IO._ftdi.SetBaudRate(460800); }			
			//if (this._isRS232) { FTDI_IO._ftdi.SetBaudRate(921600); }
			if (this.IsRS232 && baudRate > 0) { this.FTDI.SetBaudRate(baudRate); }
		}

		public override Byte[] Read(Int32 timeout, Boolean useCount, ref String errorText) {
			Double d = Core.GetNowSecondsD();
			List<Byte> result = new List<Byte>();
			while (true) {
				if (timeout != 0 && (Core.GetNowSecondsD() - d) * 1000 > timeout) {
					errorText = "FTDI.Read timeout.";
					return null;
				}
				// Get number of bytes waiting to be read
				UInt32 RxQ_Bytes = 0;
				FTDI.FT_STATUS ftStatus = this.FTDI.GetRxBytesAvailable(ref RxQ_Bytes);
				if (ftStatus != FTDI.FT_STATUS.FT_OK) {
					errorText = "FTDI.GetRxBytesAvailable failed.";
					return null;
				}
				// Read bytes waiting
				if (RxQ_Bytes == 0) {
					if (timeout == 0) {
						// No data to read
						return new Byte[] { };
					}
					// Timeout not expired, continue waiting...
					continue;
				}
				// Allocate array to recieve data
				Byte[] btRead = new Byte[RxQ_Bytes];
				UInt32 bytesRead = 0;
				ftStatus = this.FTDI.Read(btRead, RxQ_Bytes, ref bytesRead);
				if (ftStatus != FTDI.FT_STATUS.FT_OK) {
					errorText = "FTDI.Read failed.";
					return null;
				}

				// Try again, may be something else...
				if (useCount) {
					if (btRead != null) {
						for (Int32 i = 0; i < btRead.Length; i++) { result.Add(btRead[i]); }
					}
					if (result.Count == result[0] + 2) {
						return result.ToArray();
					}
					if (result.Count >= 4 && result.Count == Number.Make32(result[3], result[2], result[1], result[0]) + 2) {
						return result.ToArray();
					}
				}
				else {
					return btRead;
				}
			}
		}

		public Byte[] ReadAsIs(ref String errorText) {
			// Get number of bytes waiting to be read
			UInt32 RxQ_Bytes = 0;
			FTDI.FT_STATUS ftStatus = this.FTDI.GetRxBytesAvailable(ref RxQ_Bytes);
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.GetRxBytesAvailable failed.";
				return null;
			}

			// Read bytes waiting
			if (RxQ_Bytes == 0) { return new Byte[] { }; }
			// Allocate array to recieve data
			Byte[] btRead = new Byte[RxQ_Bytes];
			UInt32 bytesRead = 0;
			ftStatus = this.FTDI.Read(btRead, RxQ_Bytes, ref bytesRead);
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.Read failed.";
				return null;
			}
			return btRead;
		}

		public override void Write(Byte[] btWrite, ref String errorText) {
			if (btWrite.Length > 0) {
				UInt32 bytesWritten = 0;
				FTDI.FT_STATUS ftStatus = this.FTDI.Write(btWrite, btWrite.Length, ref bytesWritten);
				if (ftStatus != FTDI.FT_STATUS.FT_OK) {
					errorText = "Write failed.";
					return;
				}
				if (bytesWritten != btWrite.Length) {
					errorText = "Write failed. bytesWritten=" + bytesWritten + ", data.Length=" + btWrite.Length;
					return;
				}
			}
			errorText = "";
		}

		public override void Close(ref String errorText) {
			// Close device
			FTDI.FT_STATUS ftStatus = this.FTDI.Close();
			if (ftStatus != FTDI.FT_STATUS.FT_OK) {
				errorText = "FTDI.Close failed.";
				return;
			}
			errorText = "";
		}
	}
}